/*
 * Portux Panel-PC and MoLUX input driver
 *
 * Copyright (C) 2005-2007 taskit GmbH
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/i2c.h>
#include <linux/interrupt.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <asm/uaccess.h>
#include <linux/input.h>

#include <linux/workqueue.h>
#include <asm/semaphore.h>
#include <linux/delay.h>

#include <linux/platform_device.h>
#include <linux/poll.h>

#include <linux/portuxinput.h>

#include <asm/arch/gpio.h>

#define try(x) (x)<0 ? ((x)<0) : 0

/* I2C */
static unsigned short normal_i2c[]={0x10, I2C_CLIENT_END};
I2C_CLIENT_INSMOD;
/* prototypes */
struct portuxinput_t;
static int portuxinput_attach_adapter(struct i2c_adapter *adapter);
static int portuxinput_detach_client(struct i2c_client *client);
static long portuxinput_ioctl(struct file *filp, unsigned int cmd, unsigned long arg);
static unsigned int portuxinput_poll(struct file *filp, struct poll_table_struct *wait);
static int portuxinput_read(struct file *filp, char __user *buf, size_t count, loff_t *offp);
static int portuxinput_open(struct inode *inode, struct file *file);
static int portuxinput_release(struct inode *inode, struct file *file);
static inline int portuxinput_i2c_write_control(struct portuxinput_t *drvstruct);
static int portuxinput_i2c_read_matrix(struct portuxinput_t *drvstruct);
static int portuxinput_i2c_read_touch(struct portuxinput_t *drvstruct);
static int portuxinput_i2c_read_ir(struct portuxinput_t *drvstruct);
static int portuxinput_i2c_read_kbd(struct portuxinput_t *drvstruct);
static int portuxinput_i2c_read_battery(struct portuxinput_t *drvstruct);
static int portuxinput_i2c_write_alarm(struct portuxinput_t *drvstruct, unsigned int alarm);
static int portuxinput_i2c_shutdown(struct portuxinput_t *drvstruct, int deeppowerdown);
static int portuxinput_i2c_read_eeprom(struct portuxinput_t *drvstruct, unsigned short address);
static int portuxinput_i2c_write_eeprom(struct portuxinput_t *drvstruct, unsigned short address, unsigned char value);
static int portuxinput_i2c_read_touchdiv(struct portuxinput_t *drvstruct);
static int portuxinput_i2c_write_touchdiv(struct portuxinput_t *drvstruct, unsigned short div);
static int portuxinput_i2c_read_matrixdiv(struct portuxinput_t *drvstruct);
static int portuxinput_i2c_write_matrixdiv(struct portuxinput_t *drvstruct, unsigned short div);

static void _portuxinput_report_event(struct input_dev *inputdev, struct portuxinputevent *event, unsigned char repeated, struct portuxinput_layerstate *layer, int recurse);
static void keyarray_timeout(unsigned long unused);
static void send_ps2command(void *data);
static void irqhandler_bh(struct work_struct *work);

struct devicehandler_t {
	unsigned short event_mask;
	iochandler_t iochandler;
	eventhandler_t eventhandler;
	int(*read_func)(struct portuxinput_t*);
};

struct portuxinput_t {
	struct portuxinput_sram_t regs;
	unsigned short wakeupcontrol;
	unsigned short savedcontrol;
	unsigned char  laststatus;
	unsigned char firststatus;
	unsigned char firsteventhappened;
	char data_copy[16];

	struct i2c_driver driver;
	struct i2c_adapter *adapter;
	struct i2c_client *client;

	struct workqueue_struct *workqueue;
	struct work_struct work;
	struct semaphore semcontrol;
	struct semaphore semdata;
	struct completion complbatt;
	struct completion complevent;
	struct miscdevice miscdev;
	struct timer_list timer;
	int timeout;
	wait_queue_head_t outq;

	keyarray *keyarraymap;
	unsigned char curkeyarray;
	unsigned char curkeyindex;
	unsigned char keynumber;
	unsigned char keyarraymode;

	struct devicehandler_t devicehandler[PORTUXINPUT_EVENT_MAXHANDLER];
	
	struct portuxinput_data *platform_data;
};

static struct ps2command_t {
	unsigned char command;
	unsigned char param;
	unsigned char send;
} ps2command={0,0,0};

static struct file_operations fops={
	.unlocked_ioctl = portuxinput_ioctl,
	.read = portuxinput_read,
	.poll = portuxinput_poll,
	.open = portuxinput_open,
	.release = portuxinput_release
};

static struct portuxinput_t portuxinput={
	.regs.control=PORTUXINPUT_MASK_INTERRUPT|PORTUXINPUT_MASK_ALARM,
	.wakeupcontrol=PORTUXINPUT_MASK_INTERRUPT,
	.semcontrol=__SEMAPHORE_INIT(portuxinput.semcontrol,1),
	.semdata=__SEMAPHORE_INIT(portuxinput.semdata,1),
	.work=__WORK_INITIALIZER(portuxinput.work, irqhandler_bh),
	.complbatt=COMPLETION_INITIALIZER(portuxinput.complbatt),
	.complevent=COMPLETION_INITIALIZER(portuxinput.complevent),
	.timer=TIMER_INITIALIZER(keyarray_timeout, 0, 0),
	.timeout=10,
	.outq=__WAIT_QUEUE_HEAD_INITIALIZER(portuxinput.outq),
	.driver={
		.id=PORTUXINPUT_ID,
		.attach_adapter=&portuxinput_attach_adapter,
		.detach_client=&portuxinput_detach_client
	},
	.devicehandler={
		{PORTUXINPUT_MASK_MATRIX,NULL,NULL,&portuxinput_i2c_read_matrix},
		{PORTUXINPUT_MASK_TOUCH ,NULL,NULL,&portuxinput_i2c_read_touch},
		{PORTUXINPUT_MASK_IR	,NULL,NULL,&portuxinput_i2c_read_ir},
		{PORTUXINPUT_MASK_KBD	,NULL,NULL,&portuxinput_i2c_read_kbd}
	},
	.miscdev={
		.minor=MISC_DYNAMIC_MINOR,
		.name=PORTUXINPUT_DEVNAME,
		.fops=&fops
	},
	.curkeyarray=0,
	.curkeyindex=0xff,
	.keynumber=0xff,
	.keyarraymode=PORTUXINPUT_KEYARRAY_MODE_BACKSPACE
};

static long portuxinput_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	int i;
	int ret=0;
	unsigned char temp[16];
	keyarray temp2;

	if(_IOC_TYPE(cmd)!=PORTUXINPUT_IOC_MAGIC) return -ENOTTY;
	if(_IOC_NR(cmd)>PORTUXINPUT_IOC_MAXNR) return -ENOTTY;

	switch(cmd) {
	case PORTUXINPUT_IOC_GETTOUCHDIV:
		down(&portuxinput.semdata);
		if(portuxinput_i2c_read_touchdiv(&portuxinput))
			ret=-EIO;
		else
			ret=*(unsigned short*)portuxinput.regs.data;
		up(&portuxinput.semdata);
		return ret;
	case PORTUXINPUT_IOC_SETTOUCHDIV:
		down(&portuxinput.semdata);
		if(portuxinput_i2c_write_touchdiv(&portuxinput, (unsigned short)arg))
			ret=-EIO;
		up(&portuxinput.semdata);
		return ret;
	case PORTUXINPUT_IOC_GETMATRIXDIV:
		down(&portuxinput.semdata);
		if(portuxinput_i2c_read_matrixdiv(&portuxinput))
			ret=-EIO;
		else
			ret=*(unsigned short*)portuxinput.regs.data;
		up(&portuxinput.semdata);
		return ret;
	case PORTUXINPUT_IOC_SETMATRIXDIV:
		down(&portuxinput.semdata);
		if(portuxinput_i2c_write_matrixdiv(&portuxinput, (unsigned short)arg))
			ret=-EIO;
		up(&portuxinput.semdata);
		return ret;
	case PORTUXINPUT_IOC_SETKEYARRAYMODE:
		if((unsigned char)arg<3) {
			keyarray_timeout(portuxinput.timer.data);
			portuxinput.keyarraymode=(unsigned char)arg;
			return 0;
		} else
			return -EINVAL;
	case PORTUXINPUT_IOC_SETKEYARRAY:
		if(copy_from_user(temp2, (__user void*)arg, sizeof(keyarray)))
			return -EINVAL;
		memcpy(portuxinput.keyarraymap, temp2, sizeof(keyarray));
		return 0;
	case PORTUXINPUT_IOC_READBATTERY:
		if(arg!=0) {
			down(&portuxinput.semcontrol);
			portuxinput.regs.control|=PORTUXINPUT_MASK_BATT;
			INIT_COMPLETION(portuxinput.complbatt);
			portuxinput_i2c_write_control(&portuxinput);
			up(&portuxinput.semcontrol);
			wait_for_completion_interruptible_timeout(&portuxinput.complbatt, (unsigned int)arg);
			down(&portuxinput.semcontrol);
			portuxinput.regs.control&=~PORTUXINPUT_MASK_BATT;
			portuxinput_i2c_write_control(&portuxinput);
			up(&portuxinput.semcontrol);
		}
		down(&portuxinput.semdata);
		if(portuxinput_i2c_read_battery(&portuxinput))
			ret=-EIO;
		else
			ret=*(unsigned short *)(portuxinput.regs.data);
		up(&portuxinput.semdata);
		return ret;
	case PORTUXINPUT_IOC_SETALARM:
		down(&portuxinput.semdata);
		if(portuxinput_i2c_write_alarm(&portuxinput, (unsigned short)arg))
			ret=-EIO;
		up(&portuxinput.semdata);
		return ret;
	case PORTUXINPUT_IOC_SETWAKEUP:
		portuxinput.wakeupcontrol|=((unsigned short)arg)<<8;
		return 0;
	case PORTUXINPUT_IOC_CLEARWAKEUP:
		portuxinput.wakeupcontrol&=~(((unsigned short)arg)<<8);
		return 0;
	case PORTUXINPUT_IOC_GETWAKEUP:
		return portuxinput.wakeupcontrol>>8;
	case PORTUXINPUT_IOC_GETFIRSTSTATUS:
		return portuxinput.firststatus;
	case PORTUXINPUT_IOC_SHUTDOWN:
		down(&portuxinput.semdata);
		if(portuxinput_i2c_shutdown(&portuxinput, arg))
			ret=-EIO;
		up(&portuxinput.semdata);
		return ret;
	case PORTUXINPUT_IOC_READEEPROM:
		if(copy_from_user(&temp, (void __user *)arg, sizeof(struct eeprom_t)))
			ret=-EFAULT;
		else {
			down(&portuxinput.semdata);
			if(portuxinput_i2c_read_eeprom(&portuxinput, ((struct eeprom_t*)temp)->address))
				ret=-EIO;
			else if(copy_to_user((void __user *)(((char*)arg)+2), &portuxinput.regs.data, 1))
				ret=-EFAULT;
			up(&portuxinput.semdata);
		}
		return ret;

	case PORTUXINPUT_IOC_WRITEEEPROM:
		if(copy_from_user(&temp, (void __user *)arg, sizeof(struct eeprom_t)))
			ret=-EFAULT;
		else {
			down(&portuxinput.semdata);
			if(portuxinput_i2c_write_eeprom(&portuxinput, ((struct eeprom_t*)temp)->address, ((struct eeprom_t*)temp)->data))
				ret=-EIO;
			up(&portuxinput.semdata);
		}
		return ret;

	default:
		for(i=0;i<PORTUXINPUT_IOC_MAXHANDLER;i++)
			if(portuxinput.devicehandler[i].iochandler)
				if((ret=portuxinput.devicehandler[i].iochandler(cmd, arg))!=-ENOTTY)
					return ret;
		return -ENOTTY;
	}
}

static int portuxinput_read(struct file *filp, char __user *buf, size_t count, loff_t *offp)
{
	while(portuxinput.laststatus==0) {
		if(filp->f_flags&O_NONBLOCK)
			return -EAGAIN;
		if(wait_event_interruptible(portuxinput.outq, (portuxinput.laststatus!=0)))
			return -ERESTARTSYS;
	}
	if(copy_to_user(buf, &portuxinput.laststatus, 1))
		return -EFAULT;
	portuxinput.laststatus=0;
	return 1;
}

static unsigned int portuxinput_poll(struct file *filp, struct poll_table_struct *wait) {
	poll_wait(filp, &portuxinput.outq, wait);
	if(portuxinput.laststatus!=0)
		return POLLIN|POLLRDNORM;
	return 0;
}

static int portuxinput_open(struct inode *inode, struct file *file)
{
	try_module_get(THIS_MODULE);
	nonseekable_open(inode, file);
	return 0;
}

static int portuxinput_release(struct inode *inode, struct file *file)
{
	module_put(THIS_MODULE);
	return 0;
}

/********************************************************************************************/

static int portuxinput_i2c_read(struct portuxinput_t *drvstruct, u8 address, u8 *buf, int len)
{
	if (drvstruct->client==NULL)
		return -ENXIO;

	int ret;
	struct i2c_msg read_msg[2]={
		{drvstruct->client->addr,       0,  1,&address},
		{drvstruct->client->addr,I2C_M_RD,len,     buf}
	};
	if((ret=i2c_transfer(drvstruct->adapter, read_msg, 2))!=2) {
		ret=-ENXIO;
	}
	return ret;
}

static int portuxinput_i2c_write(struct portuxinput_t *drvstruct, u8 address, u8 *buf, int len)
{
	if (drvstruct->client==NULL)
		return -ENXIO;

	int ret;
	u8 *lbuf;
	struct i2c_msg write_msg={drvstruct->client->addr,0,len+1,NULL};

	if(IS_ERR(lbuf=kmalloc(len+1, GFP_KERNEL)))
		return -ENOMEM;

	lbuf[0]=address;
	if(len>0)
		memcpy(lbuf+1, buf, len);
	write_msg.buf=lbuf;
	if((ret=i2c_transfer(drvstruct->adapter, &write_msg, 1))!=1) {
		ret=-ENXIO;
	}
	kfree(lbuf);
	return ret;
}

static inline int portuxinput_i2c_read_status(struct portuxinput_t *drvstruct)
{
	return portuxinput_i2c_read(drvstruct, PORTUXINPUT_STATUS, (char *)&(drvstruct)->regs.status, 2);
}
static inline int portuxinput_i2c_write_control(struct portuxinput_t *drvstruct)
{
	return portuxinput_i2c_write(drvstruct, PORTUXINPUT_CONTROL, (char *)&(drvstruct)->regs.control, 2);
}
static inline int portuxinput_i2c_write_command(struct portuxinput_t *drvstruct)
{
	return portuxinput_i2c_write(drvstruct, PORTUXINPUT_COMMAND, &(drvstruct)->regs.command, 1);
}
static inline int portuxinput_i2c_read_data(struct portuxinput_t *drvstruct, int len)
{
	return portuxinput_i2c_read(drvstruct, PORTUXINPUT_DATA, (drvstruct)->regs.data, len);
}
static inline int portuxinput_i2c_write_data(struct portuxinput_t *drvstruct, int len)
{
	return portuxinput_i2c_write(drvstruct, PORTUXINPUT_DATA, (drvstruct)->regs.data, len);
}

static int portuxinput_i2c_wait_finish (struct portuxinput_t *drvstruct)
{
	int timeout=200;
	while(timeout--) {
		portuxinput_i2c_read_status(drvstruct);
		if(drvstruct->regs.status&PORTUXINPUT_MASK_FINISHED) {
			if(drvstruct->regs.status&PORTUXINPUT_MASK_ERROR)
				return -1;
			return 0;
		}
		msleep(20);
	};
	return -1;
}

/********************************************************************************************/

static int portuxinput_i2c_data_command_data(struct portuxinput_t *drvstruct, int srclen, int dstlen) {
	if(portuxinput_i2c_write_data(drvstruct, srclen)<0)
		return -EIO;
	if(portuxinput_i2c_write_command(drvstruct)<0)
		return -EIO;
	if(portuxinput_i2c_wait_finish(drvstruct)<0)
		return -EIO;
	if(portuxinput_i2c_read_data(drvstruct, dstlen)<0)
		return -EIO;
	return 0;
}

static int portuxinput_i2c_data_command(struct portuxinput_t *drvstruct, int srclen, int wait_finish) {
	if(portuxinput_i2c_write_data(drvstruct, srclen)<0)
		return -EIO;
	if(portuxinput_i2c_write_command(drvstruct)<0)
		return -EIO;
	if(wait_finish)
		if(portuxinput_i2c_wait_finish(drvstruct)<0)
			return -EIO;
	return 0;
}

static int portuxinput_i2c_command_data(struct portuxinput_t *drvstruct, int dstlen) {
	if(portuxinput_i2c_write_command(drvstruct)<0)
		return -EIO;
	if(portuxinput_i2c_wait_finish(drvstruct)<0)
		return -EIO;
	if(portuxinput_i2c_read_data(drvstruct, dstlen)<0)
		return -EIO;
	return 0;
}

/********************************************************************************************/

static int portuxinput_i2c_read_eeprom(struct portuxinput_t *drvstruct, unsigned short address)
{
	*((unsigned short*)drvstruct->regs.data)=address;
	drvstruct->regs.command=PORTUXINPUT_READEEPROM;
	return portuxinput_i2c_data_command_data(drvstruct, 2, 1);
}

static int portuxinput_i2c_write_eeprom(struct portuxinput_t *drvstruct, unsigned short address, unsigned char value)
{
	*((unsigned short*)drvstruct->regs.data)=address;
	drvstruct->regs.data[2]=value;
	drvstruct->regs.command=PORTUXINPUT_WRITEEEPROM;
	return portuxinput_i2c_data_command(drvstruct, 3, 1);
}

static int portuxinput_i2c_read_matrix(struct portuxinput_t *drvstruct)
{
	drvstruct->regs.command=PORTUXINPUT_READMATRIX;
	return portuxinput_i2c_command_data(drvstruct, sizeof(struct matrix_t));
}

static int portuxinput_i2c_read_touch(struct portuxinput_t *drvstruct)
{
	drvstruct->regs.command=PORTUXINPUT_READTOUCH;
	return portuxinput_i2c_command_data(drvstruct, sizeof(struct touch_t));
}

static int portuxinput_i2c_read_ir(struct portuxinput_t *drvstruct)
{
	drvstruct->regs.command=PORTUXINPUT_READIR;
	return portuxinput_i2c_command_data(drvstruct, sizeof(struct ir_t));
}

static int portuxinput_i2c_read_kbd(struct portuxinput_t *drvstruct)
{
	drvstruct->regs.command=PORTUXINPUT_READKBD;
	return portuxinput_i2c_command_data(drvstruct, 1);
}

static int portuxinput_i2c_write_kbd(struct portuxinput_t *drvstruct, unsigned char ps2cmd)
{
	drvstruct->regs.data[0]=ps2cmd;
	drvstruct->regs.command=PORTUXINPUT_WRITEKBD;
	return portuxinput_i2c_data_command(drvstruct, 1, 1);
}

static int portuxinput_i2c_read_touchdiv(struct portuxinput_t *drvstruct)
{
	drvstruct->regs.command=PORTUXINPUT_GETTOUCHDIV;
	return portuxinput_i2c_command_data(drvstruct, 2);
}

static int portuxinput_i2c_write_touchdiv(struct portuxinput_t *drvstruct, unsigned short div)
{
	drvstruct->regs.command=PORTUXINPUT_SETTOUCHDIV;
	*(unsigned short*)drvstruct->regs.data=div;
	return portuxinput_i2c_data_command(drvstruct, 2, 1);
}

static int portuxinput_i2c_read_matrixdiv(struct portuxinput_t *drvstruct)
{
	drvstruct->regs.command=PORTUXINPUT_GETMATRIXDIV;
	return portuxinput_i2c_command_data(drvstruct, 2);
}

static int portuxinput_i2c_write_matrixdiv(struct portuxinput_t *drvstruct, unsigned short div)
{
	drvstruct->regs.command=PORTUXINPUT_SETMATRIXDIV;
	*(unsigned short*)drvstruct->regs.data=div;
	return portuxinput_i2c_data_command(drvstruct, 2, 1);
}

static int portuxinput_i2c_read_battery(struct portuxinput_t *drvstruct)
{
	drvstruct->regs.command=PORTUXINPUT_READBATT;
	return portuxinput_i2c_command_data(drvstruct, 2);
}

static int portuxinput_i2c_write_alarm(struct portuxinput_t *drvstruct, unsigned int alarm)
{
	drvstruct->regs.command=PORTUXINPUT_SETALARM;
	*(unsigned int*)drvstruct->regs.data=alarm;
	return portuxinput_i2c_data_command(drvstruct, 4, 1);
}

static int portuxinput_i2c_shutdown(struct portuxinput_t *drvstruct, int deeppowerdown)
{
	memset(drvstruct->regs.data, 0, sizeof(drvstruct->regs.data));
	if(deeppowerdown)
		drvstruct->regs.data[0]=1;
	drvstruct->regs.command=PORTUXINPUT_SHUTDOWN;
	return portuxinput_i2c_data_command(drvstruct, sizeof(drvstruct->regs.data), 0);
}

/********************************************************************************************/

/**
 * portuxinput_report_modifier - reports modifier keys as set in @event and @value
 * @inputdev: pointer to the input device structure to use to report
 * @event: pointer to an event structure, from which the modifier keys should be uses
 * @value: determines, whether the modifier keys should be reported as pressed or released
 * Description: This is a helper function for portuxinput_report_event(). It
 * reports modifier key pressing and releasing as set in @event->flags. A value
 * of 1 means "key pressed" and 0 means "key released".
 */
static void portuxinput_report_modifier(struct input_dev *inputdev, struct portuxinputevent *event, int value)
{
	if((event->flags)&PORTUXINPUT_MASK_LSHIFT)
		input_report_key(inputdev, KEY_LEFTSHIFT, value);
	if((event->flags)&PORTUXINPUT_MASK_RSHIFT)
		input_report_key(inputdev, KEY_RIGHTSHIFT, value);
	if((event->flags)&PORTUXINPUT_MASK_LCTRL)
		input_report_key(inputdev, KEY_LEFTCTRL, value);
	if((event->flags)&PORTUXINPUT_MASK_RCTRL)
		input_report_key(inputdev, KEY_RIGHTCTRL, value);
	if((event->flags)&PORTUXINPUT_MASK_LALT)
		input_report_key(inputdev, KEY_LEFTALT, value);
	if((event->flags)&PORTUXINPUT_MASK_RALT)
		input_report_key(inputdev, KEY_RIGHTALT, value);
}

static void portuxinput_handle_keyarray(struct input_dev *inputdev, struct portuxinputevent *event)
{
	if(!portuxinput.keyarraymap)
		return;
	if(timer_pending(&portuxinput.timer))
		del_timer(&portuxinput.timer);
	if(event->value!=portuxinput.curkeyarray)
		keyarray_timeout((unsigned long)inputdev);
	portuxinput.curkeyindex++;
	portuxinput.keynumber++;
	portuxinput.curkeyarray=event->value;
	if((*portuxinput.keyarraymap)[portuxinput.curkeyarray][portuxinput.curkeyindex].value==0||
	   portuxinput.curkeyindex==PORTUXINPUT_MAX_KEYARRAYELEMENTS)
		portuxinput.curkeyindex=0;
	if(portuxinput.keyarraymode==PORTUXINPUT_KEYARRAY_MODE_BACKSPACE&&
	   (*portuxinput.keyarraymap)[portuxinput.curkeyarray][portuxinput.curkeyindex].value!=0&&
	   portuxinput.keynumber>0) {
		input_report_key(inputdev,KEY_BACKSPACE, 1);
		input_report_key(inputdev,KEY_BACKSPACE, 0);
		input_sync(inputdev);
	}
	if(portuxinput.keyarraymode!=PORTUXINPUT_KEYARRAY_MODE_BLIND) {
		_portuxinput_report_event(inputdev, &(*portuxinput.keyarraymap)[portuxinput.curkeyarray][portuxinput.curkeyindex],0,NULL,0);
	}
	portuxinput.timer.expires=jiffies+(portuxinput.timeout*HZ)/10;
	portuxinput.timer.data=(unsigned long)inputdev;
	add_timer(&portuxinput.timer);
}

static void _portuxinput_report_event(struct input_dev *inputdev, struct portuxinputevent *event, unsigned char repeated, struct portuxinput_layerstate *layer, int recurse)
{
	if(repeated&&!(event->flags&PORTUXINPUT_MASK_REPEAT))
		return;

	if(event->flags&PORTUXINPUT_MASK_KEYARRAY) {
		if(recurse)
			portuxinput_handle_keyarray(inputdev, event);
	} else {
		if(recurse)
			keyarray_timeout(portuxinput.timer.data);

		if(layer!=NULL&&event->flags&PORTUXINPUT_MASK_SETKEYMAP) {
			if(event->value>=0&&
			   event->value<layer->max)
				layer->cur=event->value;
		} else if(layer!=NULL&&event->flags&PORTUXINPUT_MASK_SETKEYMAPRET) {
			if(event->value>=0&&
			   event->value<layer->max) {
				layer->last=layer->cur;
				layer->cur=event->value;
				layer->ret=1;
			}
		} else if(event->flags&PORTUXINPUT_MASK_MOUSEMOVE) {
			if(event->flags&PORTUXINPUT_MASK_RELX)
				input_report_rel(inputdev, REL_X, event->value);
			else if(event->flags&PORTUXINPUT_MASK_RELY)
				input_report_rel(inputdev, REL_Y, event->value);
		} else {
			portuxinput_report_modifier(inputdev, event, 1);
			input_report_key(inputdev, event->value, 1);
			input_sync(inputdev);
			input_report_key(inputdev, event->value, 0);
			portuxinput_report_modifier(inputdev, event, 0);
		}
	}
	if(layer!=NULL)
		if(layer->ret&&!repeated) {
			layer->ret=0;
			layer->cur=layer->last;
		}
}

static void keyarray_timeout(unsigned long inputdev) {
	if(portuxinput.keyarraymode==PORTUXINPUT_KEYARRAY_MODE_BLIND&&portuxinput.keynumber!=0xff&&inputdev!=0)
		_portuxinput_report_event((struct input_dev*)inputdev, &(*portuxinput.keyarraymap)[portuxinput.curkeyarray][portuxinput.curkeyindex],0,NULL,0);
	portuxinput.curkeyindex=0xff;
	portuxinput.keynumber=0xff;
}

/**
 * portuxinput_report_event - reports input events to the input system
 * @inputdev: pointer to the input device structure to use to report
 * @event: event to report
 * @repeated: if non-zero, the event is marked as repeated
 * Description: This function can be used by input event handling functions
 * to report events. It can handle key/button and relative x and y movement
 * events. If @repeated is non-zero, the event is only delivered to the input
 * system, if PORTUXINPUT_MASK_REPEAT is set in event->flags.
 */
void portuxinput_report_event(struct input_dev *inputdev, struct portuxinputevent *event, unsigned char repeated, struct portuxinput_layerstate *layer)
{
	_portuxinput_report_event(inputdev, event, repeated, layer, 1);
}

/**
 * portuxinput_register_eventhandler - registers event and ioctl handler functions
 * @event_id: determines, on which event @handler should be called
 * @handler: handler, that should be called, whenever event @event_id occurs
 * @iochandler: handler for ioctl commands, can be NULL
 * Returns: zero on success, non-zero on failure
 * Description: This function registers handler funtions for events and ioctl
 * commands. It will fail, if event_id is not a valid id, another event handling
 * function is already registered, handler is NULL or the notification of the
 * device failed.
 * Note: This function should not be called directly. The wrapper functions
 * portuxinput_register_matrix(), portuxinput_register_touch() and
 * portuxinput_register_ir(), which omit the event_id parameter, should be used
 * instead. Additionally, this function may sleep.
 */
int portuxinput_register_eventhandler(int event_id, eventhandler_t handler, iochandler_t iochandler)
{
	if(event_id>=PORTUXINPUT_EVENT_MAXHANDLER)
		return -1;
	if(portuxinput.devicehandler[event_id].eventhandler||handler==NULL)
		return -1;
	down(&portuxinput.semcontrol);
	portuxinput.regs.control|=portuxinput.devicehandler[event_id].event_mask;
	portuxinput_i2c_write_control(&portuxinput);
	up(&portuxinput.semcontrol);
	portuxinput.devicehandler[event_id].eventhandler=handler;
	portuxinput.devicehandler[event_id].iochandler=iochandler;
	return 0;
}

/**
 * portuxinput_deregister_eventhandler - deregisters event and ioctl handler functions
 * @event_id: determines, on which @handler should be called
 * @handler: handler, that should be called, whenever event @event_id occurs
 * Returns: zero on success, non-zero on failure
 * Description: This function deregisters the handler funtions for events and
 * ioctl commands. It will fail, if event_id is not a valid id or handler is not
 * equal to the registered handler.
 * Note: This function should not be called directly. The wrapper functions
 * portuxinput_deregister_matrix(), portuxinput_deregister_touch() and
 * portuxinput_deregister_ir(), which omit the event_id parameter, should be
 * used instead. Additionally, this function may sleep.
 */
int portuxinput_deregister_eventhandler(int event_id, eventhandler_t handler)
{
	if(event_id>=PORTUXINPUT_EVENT_MAXHANDLER)
		return -1;
	if(portuxinput.devicehandler[event_id].eventhandler==handler) {
		portuxinput.devicehandler[event_id].eventhandler=NULL;
		portuxinput.devicehandler[event_id].iochandler=NULL;
		down(&portuxinput.semcontrol);
		portuxinput.regs.control&=~portuxinput.devicehandler[event_id].event_mask;
		portuxinput_i2c_write_control(&portuxinput);
		up(&portuxinput.semcontrol);
		return 0;
	} else
		return -1;
}

static int portuxinput_detect_client(struct i2c_adapter *adapter, int address, int kind)
{
	struct i2c_client *new_client;
	int err;
	if(!(new_client=kmalloc(sizeof(struct i2c_client), GFP_KERNEL)))
		return -ENOMEM;
	memset(new_client,0,sizeof(struct i2c_client));
	new_client->addr=address;
	new_client->adapter=adapter;
	new_client->driver=&portuxinput.driver;
	strcpy(new_client->name, PORTUXINPUT_NAME);
	if((err=i2c_attach_client(new_client))) {
		kfree(new_client);
		return err;
	}
	portuxinput.client=new_client;
	return 0;
}

static int portuxinput_attach_adapter(struct i2c_adapter *adapter)
{
	int err;
	if(!(err=i2c_probe(adapter, &addr_data, &portuxinput_detect_client))) {
		portuxinput.adapter=adapter;
		portuxinput_i2c_write_control(&portuxinput);
	}
	return err;
}

static int portuxinput_detach_client(struct i2c_client *client)
{
	int ret;
	if((ret=i2c_detach_client(client)))
		return ret;
	kfree(client);
	return 0;
}

static void send_ps2command(void *data)
{
	portuxinput_i2c_write_kbd(&portuxinput, ((struct ps2command_t*)data)->command);
	portuxinput_i2c_write_kbd(&portuxinput, ((struct ps2command_t*)data)->param);
}

/**
 * portuxinput_schedule_ps2command - schedule a ps2 command to be send to the input controller
 * @ps2cmd: command to be executed by the ps2 keyboard
 * @param: one byte parameter belonging to the command
 * Description: This functions schedules the command given in @ps2cmd and its
 * @param for sending in process context later.
 */
void portuxinput_schedule_ps2command(unsigned char ps2cmd, unsigned char param)
{
	ps2command.command=ps2cmd;
	ps2command.param=param;
	ps2command.send=1;
}

static void irqhandler_bh(struct work_struct *work)
{
	unsigned char i;
	
	if(portuxinput.client==NULL)
		return;
	down(&portuxinput.semcontrol);
	if(try(portuxinput_i2c_read_status(&portuxinput)))
		goto error;

	if(portuxinput.regs.status&0xff00) {
		if(unlikely(portuxinput.firsteventhappened==0)) {
			portuxinput.firststatus=portuxinput.regs.status>>8;
			portuxinput.firsteventhappened=1;
		} else {
			portuxinput.laststatus=portuxinput.regs.status>>8;
			wake_up_interruptible(&portuxinput.outq);
		}
		for(i=0;i<PORTUXINPUT_EVENT_MAXHANDLER;i++)
			if(portuxinput.regs.status&portuxinput.devicehandler[i].event_mask) {
				down(&portuxinput.semdata);
				if(try(portuxinput.devicehandler[i].read_func(&portuxinput))) {
					up(&portuxinput.semdata);
					goto error;
				}
				memcpy(portuxinput.data_copy, portuxinput.regs.data, 16);
				up(&portuxinput.semdata);
				if(portuxinput.devicehandler[i].eventhandler)
					portuxinput.devicehandler[i].eventhandler(portuxinput.data_copy);
			}
	}
error:
	portuxinput_i2c_write_control(&portuxinput);
	up(&portuxinput.semcontrol);
	if(ps2command.send==1) {
		down(&portuxinput.semdata);
		send_ps2command(&ps2command);
		up(&portuxinput.semdata);
		ps2command.send=0;
	}
}

static irqreturn_t irqhandler(int irq, void *dev_id)
{
	queue_work(portuxinput.workqueue, &portuxinput.work);
	return IRQ_HANDLED;
}

struct portuxinput_data *portuxinput_get_platform_data(void) {
	return portuxinput.platform_data;
}

static int portuxinput_probe(struct platform_device *pdev)
{
	struct portuxinput_data *data=pdev->dev.platform_data;
	portuxinput.workqueue=create_singlethread_workqueue(PORTUXINPUT_DEVNAME);
	if(!portuxinput.workqueue)
		return -1;
	if(request_irq(data->irq, &irqhandler, IRQF_SHARED, PORTUXINPUT_NAME, &portuxinput)) {
		destroy_workqueue(portuxinput.workqueue);
		return -1;
	}
	i2c_add_driver(&portuxinput.driver);
	misc_register(&portuxinput.miscdev);
	portuxinput.platform_data=data;
	portuxinput.keyarraymap=data->keyarraymap;
	enable_irq_wake(data->irq);
	return 0;
}

static int portuxinput_remove(struct platform_device *pdev)
{
	struct portuxinput_data *data=platform_get_drvdata(pdev);
	free_irq(data->irq, &portuxinput);
	misc_deregister(&portuxinput.miscdev);
	flush_workqueue(portuxinput.workqueue);
	destroy_workqueue(portuxinput.workqueue);
	i2c_del_driver(&portuxinput.driver);
	return 0;
}


#ifdef CONFIG_PM
static int portuxinput_suspend(struct platform_device *pdev, pm_message_t state)
{
	portuxinput.firsteventhappened=0;
	portuxinput.savedcontrol=portuxinput.regs.control;
	portuxinput.regs.control=portuxinput.wakeupcontrol;
	portuxinput_i2c_write_control(&portuxinput);
//	portuxinput_i2c_shutdown(&portuxinput, 0);
	return 0;
}

static int portuxinput_resume(struct platform_device *pdev)
{
	portuxinput_i2c_write(&portuxinput, 0, 0, 0);
	portuxinput.regs.status=0;
	if(portuxinput.firsteventhappened==0) {
		try(portuxinput_i2c_read_status(&portuxinput));
		portuxinput.firststatus=portuxinput.regs.status>>8;
		portuxinput.firsteventhappened=1;
	}
	portuxinput.regs.control=portuxinput.savedcontrol;
	portuxinput_i2c_write_control(&portuxinput);
	return 0;
}
#else
#define portuxinput_suspend	NULL
#define portuxinput_resume	NULL
#endif

static struct platform_driver portuxinput_driver = {
	.probe		= portuxinput_probe,
	.remove		= portuxinput_remove,
	.suspend	= portuxinput_suspend,
	.resume		= portuxinput_resume,
	.driver		= {
		.name	= PORTUXINPUT_DEVNAME,
	},
};

/********************************************************************************************/

static int __init portuxinput_init(void)
{
	return platform_driver_register(&portuxinput_driver);
}

static void __exit portuxinput_exit(void)
{
 	platform_driver_unregister(&portuxinput_driver);
}

EXPORT_SYMBOL(portuxinput_register_eventhandler);
EXPORT_SYMBOL(portuxinput_deregister_eventhandler);
EXPORT_SYMBOL(portuxinput_report_event);
EXPORT_SYMBOL(portuxinput_schedule_ps2command);
EXPORT_SYMBOL(portuxinput_get_platform_data);

module_init(portuxinput_init)
module_exit(portuxinput_exit)

MODULE_AUTHOR("Christian Glindkamp");
MODULE_DESCRIPTION("Input driver for the Portux Panel-PC and MoLUX");
MODULE_LICENSE("GPL v2");

