/*
 * Portux Panel-PC touch panel submodule
 *
 * Copyright (C) 2005-2007 taskit GmbH
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/input.h>
#include <linux/timer.h>
#include <asm/uaccess.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/portuxinput.h>

static struct input_dev *portuxtouch;

static struct calibration calibration = {
	.minx=150,
	.maxx=900,
	.miny=150,
	.maxy=900
};

static void timeout(unsigned long unused);

static struct timer_list timer=TIMER_INITIALIZER(&timeout, 0, 0);
spinlock_t timerlock=SPIN_LOCK_UNLOCKED;

static int start_calibration(void)
{
	if(calibration.active)
		return -1;
	calibration.minx=512;
	calibration.maxx=512;
	calibration.miny=512;
	calibration.maxy=512;
	calibration.active=1;
	return 0;
}

static int finish_calibration(void)
{
	if(!calibration.active)
		return -1;
	spin_lock_bh(&timerlock);
	if(timer_pending(&timer))
		del_timer(&timer);
	spin_unlock_bh(&timerlock);
	calibration.active=0;
	if(calibration.miny==512||calibration.maxy==512||calibration.minx==512||calibration.maxx==512)
		return -1;
	portuxtouch->absmin[ABS_X]=calibration.minx;
	portuxtouch->absmax[ABS_X]=calibration.maxx;
	portuxtouch->absmin[ABS_Y]=calibration.miny;
	portuxtouch->absmax[ABS_Y]=calibration.maxy;
	calibration.active=0;
	return 0;
}

static void timeout(unsigned long unused)
{
	finish_calibration();
}

static long iochandler(unsigned int cmd, unsigned long arg)
{
	switch(cmd) {
		case PORTUXTOUCH_IOC_CALIBRATE:
			if(start_calibration())
				return -EBUSY;
			calibration.timeout=arg;
			spin_lock_bh(&timerlock);
			timer.expires=jiffies+calibration.timeout*HZ;
			add_timer(&timer);
			spin_unlock_bh(&timerlock);
			break;
		case PORTUXTOUCH_IOC_CALIBRATESTART:
			if(start_calibration())
				return -EBUSY;
			break;
		case PORTUXTOUCH_IOC_CALIBRATEFINISH:
			if(finish_calibration())
				return -EBUSY;
			break;
		case PORTUXTOUCH_IOC_CALIBRATEABORT:
			if(!calibration.active)
				return -EBUSY;
			calibration.active=0;
			break;
		case PORTUXTOUCH_IOC_CALIBRATIONSTATUS:
			if(calibration.active)
				calibration.remaining=(timer.expires-jiffies)/HZ;
			else
				calibration.remaining=0;
			if(copy_to_user((void __user *)arg, &calibration, sizeof(struct calibration)))
				return -EFAULT;
			break;
		case PORTUXTOUCH_IOC_LOADCALIBRATION:
			if(copy_from_user(&calibration, (void __user *)arg, sizeof(struct calibration)))
				return -EFAULT;
			calibration.active=0;
			calibration.timeout=0;
			calibration.remaining=0;
			portuxtouch->absmin[ABS_X]=calibration.minx;
			portuxtouch->absmax[ABS_X]=calibration.maxx;
			portuxtouch->absmin[ABS_Y]=calibration.miny;
			portuxtouch->absmax[ABS_Y]=calibration.maxy;
			break;
		default:
			return -ENOTTY;
	}
	return 0;
}

static void touchhandler(struct touch_t *event)
{
	if(!calibration.active) {
		if(event->x==1023||event->y==1023) {
			input_report_key(portuxtouch, BTN_TOUCH, 0);
			input_report_abs(portuxtouch, ABS_PRESSURE, 0);
		} else {
			input_report_abs(portuxtouch, ABS_X, event->x);
			input_report_abs(portuxtouch, ABS_Y, event->y);
			input_sync(portuxtouch);
			input_report_key(portuxtouch, BTN_TOUCH, 1);
			input_report_abs(portuxtouch, ABS_PRESSURE, 1);
		}
		input_sync(portuxtouch);
	} else {
		if(event->x<1023&&event->y<1023) {
			if(event->x<calibration.minx)
				calibration.minx=event->x;
			else if(event->x>calibration.maxx)
				calibration.maxx=event->x;
			if(event->y<calibration.miny)
				calibration.miny=event->y;
			else if(event->y>calibration.maxy)
				calibration.maxy=event->y;
			spin_lock_bh(&timerlock);
			del_timer(&timer);
			timer.expires=jiffies+calibration.timeout*HZ;
			add_timer(&timer);
			spin_unlock_bh(&timerlock);
		}
	}
}

static int portuxtouch_open(struct input_dev *dev)
{
	try_module_get(THIS_MODULE);
	return 0;
}

static void portuxtouch_close(struct input_dev *dev)
{
	module_put(THIS_MODULE);
}

static int __init portuxtouch_init(void)
{
	if(portuxinput_get_platform_data()) {
		portuxtouch=input_allocate_device();
		portuxtouch->name="Portux touch panel";
		portuxtouch->open=portuxtouch_open;
		portuxtouch->close=portuxtouch_close;
		portuxtouch->evbit[0]=BIT(EV_KEY)|BIT(EV_ABS);
		portuxtouch->absbit[0]=BIT(ABS_X)|BIT(ABS_Y);
		set_bit(BTN_TOUCH, portuxtouch->keybit);
		input_set_abs_params(portuxtouch, ABS_X, calibration.minx, calibration.maxx, 20, 0);
		input_set_abs_params(portuxtouch, ABS_Y, calibration.miny, calibration.maxy, 20, 0);
		input_set_abs_params(portuxtouch, ABS_PRESSURE, 0, 1, 0, 0);
		input_register_device(portuxtouch);

		portuxinput_register_touch(&touchhandler, &iochandler);
	}
	return 0;
}

static void __exit portuxtouch_exit(void)
{
	portuxinput_deregister_touch(&touchhandler);
	input_unregister_device(portuxtouch);
}

module_init(portuxtouch_init)
module_exit(portuxtouch_exit)

MODULE_AUTHOR("Christian Glindkamp");
MODULE_DESCRIPTION("touch panel submodule for portuxinput");
MODULE_LICENSE("GPL v2");

