/*
 * Family driver for the Dallas 1-Wire RTC DS2417
 *
 * Copyright (C) taskit Rechnertechnik GmbH
 *
 * Note: This driver depends on the Arm port of Linux and assumes
 * to be connected to IRQ0 on an AT91RM9200 board
 */
#include <asm/types.h>

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/device.h>
#include <linux/types.h>
#include <linux/rtc.h>
#include <asm/rtc.h>
#include <asm/arch/at91rm9200.h>

#include "../w1.h"
#include "../w1_int.h"
#include "../w1_family.h"

#define READ_CLOCK	0x66
#define WRITE_CLOCK	0x99
#define RTC_IRQ		AT91RM9200_ID_IRQ0
#define DS2417_NAME	"ds2417"

static unsigned char ds2417_ctrlreg=0x0c;

static struct w1_slave *ds2417_slave=NULL;

static int w1_ds2417_write_time(struct w1_slave *slave, struct rtc_time *tm)
{
	char data[5];
	struct w1_master *master = slave->master;
	int ret=0;

	data[0]=ds2417_ctrlreg;	
	atomic_inc(&slave->refcnt);
	smp_mb__after_atomic_inc();
	mutex_lock(&master->mutex);

	if(!w1_reset_bus(master)) {
		u8 match[9]={W1_MATCH_ROM,};
		memcpy(&match[1], &slave->reg_num, 8);
		w1_write_block(master, match, 9);
		w1_write_8(master, WRITE_CLOCK);

		rtc_tm_to_time(tm, (long*)&data[1]);
		w1_write_block(master, data, 5);
		w1_reset_bus(master);
	} else
		ret=-EIO;

	mutex_unlock(&master->mutex);
	smp_mb__before_atomic_inc();
	atomic_dec(&slave->refcnt);
	return ret;
}

static int w1_ds2417_read_time(struct w1_slave *slave, struct rtc_time *tm)
{
	char data[5];
	struct w1_master *master = slave->master;
	int ret=0;
	
	atomic_inc(&slave->refcnt);
	smp_mb__after_atomic_inc();
	mutex_lock(&master->mutex);
	
	if(!w1_reset_bus(master)) {
		u8 match[9]={W1_MATCH_ROM,};
		memcpy(&match[1], &slave->reg_num, 8);
		w1_write_block(master, match, 9);
		w1_write_8(master, READ_CLOCK);

		if(w1_read_block(master, data, 5)!=5) {
			printk(KERN_ERR "Error while reading time from DS2417.\n");
			ret=-EIO;
			goto out;
		}
		rtc_time_to_tm(*(long*)&data[1], tm);
		w1_reset_bus(master);
	} else
		ret=-EIO;
out:
	mutex_unlock(&master->mutex);
	smp_mb__before_atomic_inc();
	atomic_dec(&slave->refcnt);
	return ret;
}

static int w1_ds2417_write_ctrlreg(struct w1_slave *slave)
{
	struct rtc_time tm;
	int ret=0;
	
	if((ret=w1_ds2417_read_time(ds2417_slave, &tm))==0)
		ret=w1_ds2417_write_time(ds2417_slave, &tm);
	return ret;
}

static irqreturn_t ds2417_interrupt(int irq, void *dev_id)
{
	rtc_update(1,0);
	return IRQ_HANDLED;
}

static int ds2417_read_time(struct rtc_time *tm)
{
	return w1_ds2417_read_time(ds2417_slave, tm);
}

static int ds2417_set_time(struct rtc_time *tm)
{
	return w1_ds2417_write_time(ds2417_slave, tm);
}

static int ds2417_ioctl(unsigned int cmd, unsigned long arg)
{
	int ret=0;
	
	switch(cmd) {
	case RTC_UIE_OFF:
	case RTC_PIE_OFF:
		ds2417_ctrlreg=0x0c;
		if((ret=w1_ds2417_write_ctrlreg(ds2417_slave))==0);
			free_irq(RTC_IRQ, NULL);
		break;
	case RTC_UIE_ON:
	case RTC_PIE_ON:
		set_irq_type(RTC_IRQ, IRQF_TRIGGER_FALLING);
		if((ret=request_irq(RTC_IRQ, ds2417_interrupt, IRQF_DISABLED, DS2417_NAME, NULL)))
			break;
		ds2417_ctrlreg=0x8c;
		ret=w1_ds2417_write_ctrlreg(ds2417_slave);
		break;
	case RTC_IRQP_READ:
		ret=put_user(1, (unsigned long *)arg);
		break;
	case RTC_IRQP_SET:
		if(arg!=1)
			ret=-EINVAL;
		break;
	default:
		ret=-EINVAL;
	}
	return ret;
}

static struct rtc_ops ds2417_rtc={
	.owner=THIS_MODULE,
	.read_time=&ds2417_read_time,
	.set_time=&ds2417_set_time,
	.ioctl=&ds2417_ioctl
};

static int w1_ds2417_add_slave(struct w1_slave *slave)
{
	if(ds2417_slave==NULL)
		if(!register_rtc(&ds2417_rtc))
			ds2417_slave=slave;
	return 0;
}

static void w1_ds2417_remove_slave(struct w1_slave *slave)
{
	if(ds2417_slave==slave) {
		unregister_rtc(&ds2417_rtc);
		ds2417_slave=NULL;
	}
}

static struct w1_family_ops w1_ds2417_fops = {
	.add_slave=&w1_ds2417_add_slave,
	.remove_slave=&w1_ds2417_remove_slave
};

static struct w1_family w1_ds2417_family = {
	.fid = 0x27,
	.fops = &w1_ds2417_fops,
};

static int __init w1_ds2417_init(void)
{
	return w1_register_family(&w1_ds2417_family);
}

static void __exit w1_ds2417_exit(void)
{
	w1_unregister_family(&w1_ds2417_family);
}

module_init(w1_ds2417_init);
module_exit(w1_ds2417_exit);

MODULE_AUTHOR("Christian Glindkamp");
MODULE_DESCRIPTION("Family driver for the Dallas 1-Wire RTC DS2417");
MODULE_LICENSE("GPL v2");
